"""
cspyce/__init__.py

CSPYCE OVERVIEW

cspyce is a Python module that provides an interface to the CSPICE library. It
implements the most widely-used functions of CSPICE in a Python-like way. It
also supports numerous enhancements, including support for Python exceptions,
array inputs, and aliases.

This version of the library has been built for the CSPICE toolkit v. 67.

cspyce contains Python interfaces to all CSPICE functions likely to be useful
to a Python programmer. Excluded are deprecated functions and various low-level
functions supporting character strings, cells, sets, file I/O, and also all
low-level "geometry finder" functions that can only be implemented in C. It
also includes vectorized versions (with suffix "_vector") for nearly every
function that receives floating-point input and does not perform I/O.

ADDED FEATURES

DOCSTRINGS
- All cspyce functions have informative docstrings, so typing
      help(function)
  provides useful information. However, the call signature appearing in the
  first line of the help text is still defined by cspyce0 and may not make
  sense to the reader. This issue is fixed by module cspyce2.

DEFAULTS
- Many cspyce functions take sensible default values if input arguments are
  omitted.
  - In gcpool, gdpool, gipool, and gnpool, start values default to 1.
  - The functions that take "SET" or "GET" as their first argument (erract,
    errdev, errprt, and timdef) have simplified calling options, which are
    summarized in their docstrings.

RUNTIME ERROR HANDLING

In the CSPICE error handling mechanism, the programmer must check the value
of function failed() regularly to determine if an error has occurred. However,
Python's exception handling mechanism obviates the need for this approach. In
cspyce1, all CSPICE errors raise Python exceptions.

In CSPICE, the programmer can control how C errors are handled using the
function erract(). Options include "IGNORE", "REPORT", "ABORT", "DEFAULT",
and "RETURN", In cspyce1, the "IGNORE" and "REPORT" options are disabled,
because they can leave behind corrupted memory. In interactive Python, the
"ABORT" and "DEFAULT" options are also disabled, because aborting an
interactive session would be pointless. The "ABORT" and "DEFAULT" options are
still available, though not recommended, when running programs
non-interactively.

The cspyce1 module supports two additional error actions, which are variants
on "RETURN". These are "EXCEPTION" and "RUNTIME". The only difference between
them is that "RUNTIME" consistently raises RuntimeError exceptions, whereas
"EXCEPTION" tailors the type of the exception to the situation.

HANDLING OF ERROR FLAGS

Many CSPICE functions bypass the library's own error handling mechanism;
instead they return a status flag, sometimes called "found" or "ok", or else
an empty response might indicate failure. The cspyce module provides
alternative options for these functions.

Within cspyce1, functions that return error flags have an alternative
implementation with a suffix of "_error", which uses cspyce1's Python
exception handling instead. For example, bodn2c(name) is the function that
returns two values given the name of a body, its body ID and a True/False flag
indicating whether the name was recognized. bodn2c_error() instead just
returns a single value, the body ID. However, it raises a Python exception
(KeyError or RuntimeError, depending on the erract setting) if the name is not
recognized.

The cspyce1 module provides several ways to control which version of the
function to use:

- The function use_flags() takes a function name or list of names and
  designates the original version of each function as the default. If the
  input argument is missing, _flag versions are selected universally.  With
  this option, for example, a call to cspyce.bodn2c() will actually call
  cspyce1.bodn2c_flag().

- The function use_errors() takes a function name or list of names and
  designates the _error version of each function as the default. If the input
  argument is missing, _error versions are selected universally. With this
  option, for example, a call to cspyce1.bodn2c() will actually call
  cspyce1.bodn2c_error() instead.

You can also choose between the "flag" and "error" versions of a function
using cspyce function attributes, as discussed below.

FUNCTION ATTRIBUTES

Like any other Python class, functions can have attributes. These are used to
simplify the choices of function options in cspyce1. Every cspyce1 function
has these attributes:

  error   = the version of this function that raises exceptions intead of
            returning flags.
  flag    = the version that returns flags instead of raising exceptions.
  vector  = the vectorized version of this function.
  scalar  = the un-vectorized version of this function.

If a particular option is not relevant to a function, the attribute still
exists, and instead simply returns the function itself. This makes it trivial
to choose a particular combination of features for a particular function call.
For example:
  ckgpav.vector()         same as ckgpav_vector()
  ckgpav.vector.flag()    same as ckgpav_vector()
  ckgpav.vector.error()   same as ckgpav_vector_error()
  ckgpav.error.scalar()   same as ckgpav_error()
  ckgpav.flag()           same as ckgpav()
  bodn2c.vector()         same as bodn2c()
  bodn2c.flag()           same as bodn2c()
"""

import inspect
import sys
import os

PYTHON2 = sys.version_info[0] < 3

# We allow the import of cspyce2 to fail because, during development, there are
# times when cspyce2.py might be invalid. If that happens, we still want to be
# able to work inside the cspyce directory tree. Without allowing this
# exception, it could become impossible to work on and test cspyce functions
# inside this directory tree. This should never occur during a normal run.

try:
    from .cspyce2 import *
except ImportError as err:
    if os.getenv("CSPICE_DEVELOPMENT"):
        print("Package cspyce2 not found.  Error ignored.")
    else:
        print("Set environment variable 'CSPICE_DEVELOPMENT' to '1' to ignore this error")
        raise err

# A set of keywords listing options set globally across the cspyce functions
GLOBAL_STATUS = set()

################################################################################
# Define functions to select between error and flag versions of functions
################################################################################

def use_errors(*funcs):
    """Switch the listed functions or names of functions to use the "error"
    version by default. If the list is empty, apply this operation to all cspyce
    functions.

    Note that this operation applies to all versions of each function. Versions
    are cspyce functions with the same base name, before any suffixes.
    """

    global GLOBAL_STATUS
    if funcs:
        GLOBAL_STATUS.add('ERRORS')
        GLOBAL_STATUS.discard('FLAGS')
    else:
        GLOBAL_STATUS.discard('ERRORS')
        GLOBAL_STATUS.discard('FLAGS')

    for name in _get_func_names(funcs):
        # <name>_flag must always point to the flag version
        # <name>_error must always point to the error version
        # Only function names that don't specify one of these are modified to
        # point to the error version.
        if 'error' not in name and 'flag' not in name:
            globals()[name] = globals()[name].error

def use_flags(*funcs):
    """Switch the listed functions or names of functions to use the "flag"
    version by default. If the list is empty, apply this operation to all cspyce
    functions.

    Note that this operation applies to all versions of each function. Versions
    are cspyce functions with the same base name, before any suffixes.
    """

    global GLOBAL_STATUS
    if funcs:
        GLOBAL_STATUS.discard('ERRORS')
        GLOBAL_STATUS.add('FLAGS')
    else:
        GLOBAL_STATUS.discard('ERRORS')
        GLOBAL_STATUS.discard('FLAGS')

    for name in _get_func_names(funcs):
        # <name>_flag must always point to the flag version
        # <name>_error must always point to the error version
        # Only function names that don't specify one of these are modified to
        # point to the flag version.
        if 'error' not in name and 'flag' not in name:
            globals()[name] = globals()[name].flag

################################################################################
# Define functions to select between vector and scalar versions of functions
################################################################################

def use_vectors(*funcs):
    """Switch the listed functions or names of functions to use the "vector"
    version by default. If the list is empty, apply this operation to all cspyce
    functions.

    Note that this operation applies to all versions of each function. Versions
    are cspyce functions with the same base name, before any suffixes.
    """

    global GLOBAL_STATUS
    if funcs:
        GLOBAL_STATUS.discard('ARRAYS')
        GLOBAL_STATUS.add('VECTORS')
        GLOBAL_STATUS.discard('SCALARS')
    else:
        GLOBAL_STATUS.discard('ARRAYS')
        GLOBAL_STATUS.discard('VECTORS')
        GLOBAL_STATUS.discard('SCALARS')

    for name in _get_func_names(funcs):
        # <name>_scalar must always point to the scalar version
        # <name>_vector must always point to the vector version
        # <name>_array must always point to the array version
        # Only function names that don't specify one of these are modified to
        # point to the vector version.
        if ('scalar' not in name and 'vector' not in name
                                 and 'array' not in name):
            globals()[name] = globals()[name].vector

def use_scalars(*funcs):
    """Switch the named functions, or else all relevant cspyce functions, to
    return flags instead of raising exceptions.

    Note that this operation applies to all versions of each function. Versions
    are cspyce functions with the same base name, before any suffixes.
    """

    global GLOBAL_STATUS
    if funcs:
        GLOBAL_STATUS.discard('ARRAYS')
        GLOBAL_STATUS.discard('VECTORS')
        GLOBAL_STATUS.add('SCALARS')
    else:
        GLOBAL_STATUS.discard('ARRAYS')
        GLOBAL_STATUS.discard('VECTORS')
        GLOBAL_STATUS.discard('SCALARS')

    for name in _get_func_names(funcs):
        # <name>_scalar must always point to the scalar version
        # <name>_vector must always point to the vector version
        # <name>_array must always point to the array version
        # Only function names that don't specify one of these are modified to
        # point to the scalar version.
        if ('scalar' not in name and 'vector' not in name
                                 and 'array' not in name):
            globals()[name] = globals()[name].scalar

def _get_func_names(funcs=(), source=None):
    """Convert a list of cspyce functions or names to a set of unique names,
    including all version suffixes.
    """

    source = source or globals()

    if funcs:
        validated = set()
        for func in funcs:
            # Convert names to funcs, assemble all versions
            validated |= set(get_all_versions(func, source).keys())
    else:
        validated = set(get_all_funcs(source).keys())

    return validated

################################################################################
# Functions to track down cspyce functions
################################################################################

def get_all_funcs(source=None, cspyce_dict=None):
    """Return a dictionary of all cspyce functions, keyed by their names.

    Inputs:
        source      the dictionary to search, which defaults to globals().
        cspyce_dict is used internally for recursion; it should not be
                    referenced in external calls.
    """

    source = source or globals()

    if cspyce_dict is None:
        cspyce_dict = {}

    # Add names from this source
    names = source.keys()
    for name in names:
        func = source[name]
        if not callable(func):
            continue
        if not hasattr(func, 'SIGNATURE'):
            continue

        # Stop if this function was already found; break infinite recursion
        if func.__name__ in cspyce_dict:
            continue

        # Add this function to the dictionary
        cspyce_dict[func.__name__] = func

        # Use the internal dictionary of this function as a recursive source
        _ = get_all_funcs(func.__dict__, cspyce_dict)

    return cspyce_dict

def get_all_versions(func, source=None):
    """Return a dictionary of all cspyce functions associated with this one,
    keyed by their names.

    Inputs:
        func        a cspyce function or the name of a cspyce function.
        source      the dictionary to search if func is specified by name;
                    default is globals().
    """

    func = validate_func(func, source)
    return get_all_funcs(func.__dict__)

def validate_func(func, source=None):
    """Return a cspyce function, given either its name or the function itself.
    Otherwise, raise an exception.

    Inputs:
        func        a cspyce function or the name of a cspyce function.
        source      the dictionary to search if func is specified by name;
                    default is globals().
    """

    if isinstance(func, str):
        full_name = func
        short_name = full_name.split('_')[0]
        source = source or globals()
        try:
            func = source[short_name]
        except KeyError:
            raise KeyError('Unrecognized function name "%s"' % full_name)

    if not callable(func):
        raise ValueError('Not a function: "%s"' % func.__name__)

    if not hasattr(func, 'SIGNATURE'):
        raise ValueError('Not a cspyce function: "%s"' % func.__name__)

    return func

################################################################################
# Add return annotations to all cspyce functions if this is Python 3
################################################################################

if not PYTHON2:
    for func in get_all_funcs().values():
        retnames = func.RETNAMES
        if retnames:
            sig = inspect.signature(func)
            if len(func.RETNAMES) == 1:
                func.__signature__ = sig.replace(return_annotation=retnames[0])
            else:
                func.__signature__ = sig.replace(return_annotation=retnames)

################################################################################

# This is the default at initialization
use_errors()

################################################################################



################################################################################
# cspyce/array_suppport.py
# Used internally by cspyce; not intended for direct import.
################################################################################

from __future__ import print_function

import numpy as np
import sys
import inspect
import warnings

import cspyce
import cspyce.cspyce1 as cspyce1
from cspyce.alias_support import alias_version

PYTHON2 = sys.version_info[0] < 3

# This isn't how we want to handle a ragged array
warnings.filterwarnings('ignore', category=np.VisibleDeprecationWarning)

################################################################################
# cspyce array function wrapper
################################################################################

ARRAY_NOTE = """
This version supports array inputs in place of any floating-point inputs. Array
shapes are broadcasted following NumPy rules. This function vectorizes the call
so that any iteration is performed inside C code and is therefore faster than
Python iteration.
"""

def _array_name(name):
    return name.replace('_vector', '_array')

def _arrayize_arglist(arglist):
    arrayized = []
    for arg in arglist:
        arrayized.append(arg.replace('[_', '[...'))

    return arrayized

def array_version(func):
    """Create and return the array version of a cspyce function.

    The array version is a wrapper function that applies NumPy broadcasting
    rules to the inputs of a vector function. If the array version already
    exists, just return it. If there is no vector version, the array version is
    simply the scalar version.
    """

    if hasattr(func, 'array'):
        return func.array

    # Handle vector-free functions quickly
    if '_vector' not in func.__name__ and func.vector is func:
        func.array = func
        return func

    # Apply vector in front of alias option, via recursion
    # Note: repair alias links after all the array functions are defined
    if '_alias' in func.__name__:
        noalias_array_func = array_version(func.noalias)
        alias_array_func = alias_version(noalias_array_func)

        alias_array_func.array = alias_array_func
        func.array = alias_array_func
        return alias_array_func

    # Handle scalar versions of vector functions via recursion
    if '_vector' not in func.__name__:
        array_func = array_version(func.vector)
        func.array = array_func
        return func

    # At this point, we always have the vectorized version of the function, and
    # the array version has not yet been constructed.

    # Identify the names and locations of arguments that can be vectorized;

    # INPUT_ITEMS is a dictionary indicating the required shape of each
    # floating-point input. It is keyed by both the name and index of the input
    # argument. If the length along a particular axis is unspecified (indicated
    # by "*" in the signature), the value in the tuple is zero.

    # RETURN_ITEMS is a list containing the shapes of the returned items, in
    # order.

    if not hasattr(func, 'INPUT_ITEMS'):
        input_items = {}
        for k, sig in enumerate(func.SIGNATURE):
            parts = sig.split('[')
            if parts[0] in ('int', 'bool', 'string', 'body_name', 'body_code',
                            'frame_name', 'frame_code'):
                continue

            if len(parts) == 1:
                shape = ()
            else:
                dims = (parts[1][:-1].replace('_,', '')
                                     .replace('_', '')
                                     .replace('*', '0'))
                if dims:
                    shape = tuple([int(d) for d in dims.split(',')])
                else:
                    shape = ()

            input_items[k] = shape
            input_items[func.ARGNAMES[k]] = shape

        func.INPUT_ITEMS = input_items

        return_items = []
        for sig in func.RETURNS:
            parts = sig.split('[')
            if len(parts) == 1:
                shape = ()
            else:
                dims = (parts[1][:-1].replace('_,', '')
                                     .replace('_', '')
                                     .replace('*', '0'))
                if dims:
                    shape = tuple([int(d) for d in dims.split(',')])
                else:
                    shape = ()

            return_items.append(shape)

        func.RETURN_ITEMS = return_items

    if not func.INPUT_ITEMS:
        func.array = func.vector
        return func

    def wrapper(*args, **keywords):
        return _exec_with_broadcasting(func, *args, **keywords)

    # Copy type info but not function version links
    for (key, value) in func.__dict__.items():
        if not callable(value):
            wrapper.__dict__[key] = value

    # After copy, revise SIGNATURE, RETURNS, and NOTES
    wrapper.SIGNATURE = _arrayize_arglist(func.SIGNATURE)
    wrapper.RETURNS   = _arrayize_arglist(func.RETURNS)
    wrapper.NOTES     = [ARRAY_NOTE] + wrapper.NOTES[1:] # replace vector note

    # Save key attributes of the wrapper function before returning
    cspyce1.assign_docstring(wrapper)
    wrapper.__name__ = _array_name(func.__name__)
    if PYTHON2:
        wrapper.func_defaults = func.func_defaults
    else:
        wrapper.__defaults__ = func.__defaults__
        wrapper.__signature__ = inspect.signature(func)

    # Insert mutual links
    wrapper.array  = wrapper
    wrapper.vector = func
    wrapper.scalar = func.scalar
    func.array = wrapper

    return wrapper

def _getarg(indx, args, keywords):
    if isinstance(indx, int):
        return args[indx]
    else:
        return keywords[indx]

def _setarg(indx, value, args, keywords):
    if isinstance(indx, int):
        args[indx] = value
    else:
        keywords[indx] = value

def _exec_with_broadcasting(func, *args, **keywords):
    """Main function to broadcast together the shapes of the input arguments
    and return results with the broadcasted shape, given the vectorized form of
    a function.
    """

    args = list(args)   # args must be mutable

    # Identify arguments needing broadcasting, convert to arrays
    array_args = []     # list of tuples (index or key of arg, array, rank)
    shapes = []
    for indx in list(range(len(args))) + list(keywords.keys()):

        item = func.INPUT_ITEMS.get(indx, None)
        if item is None:
            continue

        rank = len(item)

        # Get argument
        arg = _getarg(indx, args, keywords)

        # Convert to floating-point array
        error = False
        try:
            arg = np.asfarray(arg)
        except ValueError:
            error = True

        if error or arg.dtype == np.dtype('O'): # indicates a ragged array shape
            name = indx if isinstance(indx,str) else func.ARGNAMES[indx]
            cspyce1.chkin(func.array.__name__)
            cspyce1.setmsg('Ragged input array for "%s": ' % name)
            cspyce1.sigerr('SPICE(INVALIDARRAYSHAPE)')
            cspyce1.chkout(func.array.__name__)
            return None

        # Extract leading shape of each array
        if rank == 0:
            shape = arg.shape
            invalid_shape = False
        elif rank > len(arg.shape):
            shape = ()
            invalid_shape = True
        else:
            shape = arg.shape[:-rank]
            arg_item = arg.shape[-rank:]
            invalid_shape = any([d1 != 0 and d1 != d2
                                 for (d1,d2) in zip(arg_item, item)])

        if invalid_shape:
            name = indx if isinstance(indx,str) else func.ARGNAMES[indx]
            required_shape = str(item).replace('0','*').replace(',)',')')
            cspyce1.chkin(func.array.__name__)
            cspyce1.setmsg('Invalid array shape %s ' % str(arg.shape) +
                           'for input "%s" ' % name +
                           'in module %s: ' % func.__name__ +
                           '(...,%s is required' % required_shape[1:])
            cspyce1.sigerr('SPICE(INVALIDARRAYSHAPE)')
            cspyce1.chkout(func.array.__name__)
            return None

        # Unit-sized items never need broadcasting
        if all([d == 1 for d in shape]):
            continue

        shapes.append(shape)
        array_args.append((indx, arg, rank))

    # Call function now if iteration is not needed
    if not array_args:
        return func.vector.__call__(*args, **keywords)

    # Determine the broadcasted shape
    try:
        broadcasted_shape = np.broadcast_shapes(*shapes)
    except ValueError:
        cspyce1.chkin(func.array.__name__)
        cspyce1.setmsg('Incompatible shapes for broadcasting: ' +
                       str(shapes)[1:-1])
        cspyce1.sigerr('SPICE(ARRAYSHAPEMISMATCH)')
        cspyce1.chkout(func.array.__name__)
        return None

    # Broadcast each array and prepare for vectorization
    for (indx, arg, rank) in array_args:

        # Determine required shape
        if rank:
            item = arg.shape[-rank:]
            required_shape = broadcasted_shape + item
        else:
            item = ()
            required_shape = broadcasted_shape

        # Leading axes are handled by iteration through the vector and can be
        # ignored
        required_shape = required_shape[-len(arg.shape):]

        # Leading unit axes can also be ignored
        arg_shape = arg.shape
        for d in arg_shape:
            if d == 1:
                required_shape = required_shape[1:]
                arg_shape = arg_shape[1:]
                arg = arg.reshape(arg.shape)
            else:
                break

        # Broadcast and copy if necessary
        if arg_shape == required_shape:
            arg = np.ascontiguousarray(arg)
        else:
            arg = np.broadcast_to(arg, required_shape).copy()

        # Flatten the leading axes
        arg = arg.reshape((-1,) + item)

        # Restore into the function arguments
        _setarg(indx, arg, args, keywords)

    # Execute the function
    cspyce1.chkin(func.array.__name__)
    results = func.__call__(*args, **keywords)

    if cspyce1.failed():
        cspyce1.chkout(func.array.__name__)
        return None

    # Reshape the results
    multiple_results = isinstance(results, list)
    if not multiple_results:
        results = [results]

    for indx,result in enumerate(results):
        rank = len(func.RETURN_ITEMS[indx])
        if rank:
            result = result.reshape(broadcasted_shape + result.shape[-rank:])
        else:
            result = result.reshape(broadcasted_shape)

        results[indx] = result

    # Return results
    cspyce1.chkout(func.array.__name__)

    if multiple_results:
        return results
    else:
        return results[0]

################################################################################
# Define the alias function selector
################################################################################

SPYCE_DICT = cspyce.__dict__

def use_arrays(*funcs):
    """Switch the listed functions or names of functions to use the "array"
    version by default. This affects all versions of any given cspyce function.
    If the list is empty, apply this operation to all cspyce functions.
    """

    if funcs:
        cspyce.GLOBAL_STATUS.add('ARRAYS')
        cspyce.GLOBAL_STATUS.discard('VECTORS')
        cspyce.GLOBAL_STATUS.discard('SCALARS')
    else:
        cspyce.GLOBAL_STATUS.discard('ARRAYS')
        cspyce.GLOBAL_STATUS.discard('VECTORS')
        cspyce.GLOBAL_STATUS.discard('SCALARS')

    for name in cspyce._get_func_names(funcs, source=SPYCE_DICT):
        # <name>_scalar must always point to the scalar version
        # <name>_vector must always point to the vector version
        # <name>_array must always point to the array version
        # Only function names that don't specify one of these are modified to
        # point to the array version.
        if ('scalar' not in name and 'vector' not in name
                                 and 'array' not in name):
            SPYCE_DICT[name] = SPYCE_DICT[name].array

################################################################################
# Function to define array versions and links for all cspyce functions
################################################################################

def _define_all_array_versions():
    """Generate all missing array functions and set the "array" attributes for
    all cspyce functions.

    This routine can be run multiple times. At each run, it only creates
    whatever is missing.
    """

    # Define an _array function for each cspyce _vector function
    avpairs = []
    funcs = cspyce.get_all_funcs(SPYCE_DICT).values()

    # Do non-alias versions first; otherwise some .array attributes are broken
    for func in funcs:
        if 'alias' in func.__name__:
            continue

        afunc = array_version(func)
        if afunc is func:               # function could not use arrays
            continue

        aname = afunc.__name__
        SPYCE_DICT[aname] = afunc

        if '_vector' in func.__name__:
            avpairs.append((afunc, func))

    for func in funcs:
        if 'alias' not in func.__name__:
            continue

        afunc = array_version(func)
        if afunc is func:               # function could not use arrays
            continue

        aname = afunc.__name__
        SPYCE_DICT[aname] = afunc

        if '_vector' in func.__name__:
            avpairs.append((afunc, func))

    # At this point, every vector function has an array function counterpart.
    # Also, every function has an "array" attribute, which points to itself if
    # there is no array version. However, the new array functions still need
    # internal links to other associated versions.

    # Fill in missing version attributes for every array function
    for (afunc, func) in avpairs:
        _define_missing_versions_for_array(afunc, func)

def _define_missing_versions_for_array(afunc, vfunc):
    """Complete the missing version attributes for a given array function."""

    for (key, version) in vfunc.__dict__.items():
        if not callable(version):
            continue
        if key in ('vector', 'scalar', 'array'):
            afunc.__dict__[key] = version
        else:
            version_name = version.__name__.replace('_vector', '_array')
            afunc.__dict__[key] = SPYCE_DICT[version_name]

################################################################################



################################################################################
# cspyce/cspyce1_info.py
#
# A dictionary of docstrings and call signatures, keyed by the name of the
# CSPICE function name.
#
# This function reads the associated cspyce0 dictionaries and updates them for
# the cspyce1 "_error" versions.
#
# Used internally by cspyce; not intended for direct import.
################################################################################

from cspyce.cspyce0_info import \
    CSPYCE_SIGNATURES, CSPYCE_ARGNAMES, CSPYCE_DEFAULTS, \
    CSPYCE_RETURNS, CSPYCE_RETNAMES, CSPYCE_DEFINITIONS, \
    CSPYCE_ABSTRACT, CSPYCE_PS, CSPYCE_URL

ERROR_INFO = [      # (function name, return value to remove if any, postscript text for docstring)

# Functions defined in cspyce0.i
    ('bodc2n', 'found' , 'Raise KeyError(BODYIDNOTFOUND) if name could not be translated.'),
    ('bodn2c', 'found' , 'Raise KeyError(BODYNAMENOTFOUND) if name could not be translated.'),
    ('bods2c', 'found' , 'Raise KeyError(BODYNAMENOTFOUND) if name could not be translated.'),
    ('ccifrm', 'found' , 'Raise ValueError(INVALIDFRAMEDEF) if frame is not found.'),
    ('cidfrm', 'found' , 'Raise KeyError(BODYIDNOTFOUND) if the requested frame information is unavailable.'),
    ('ckcov' , ''      , 'Raise KeyError(BODYIDNOTFOUND) if the body code is not found in the C kernel.'),
    ('ckgp'  , 'found' , 'Raise IOError(CKINSUFFDATA) if the requested pointing is unavailable.'),
    ('ckgpav', 'found' , 'Raise IOError(CKINSUFFDATA) if the requested pointing is unavailable.'),
    ('cnmfrm', 'found' , 'Raise KeyError(BODYNAMENOTFOUND) if the requested frame information is unavailable.'),
    ('dtpool', 'found' , 'Raise KeyError(VARIABLENOTFOUND) if the requested variable is not in the kernel pool.'),
    ('frinfo', 'found' , 'Raise KeyError(FRAMEIDNOTFOUND) if the requested frame is not found.'),
    ('frmnam', ''      , 'Raise KeyError(FRAMEIDNOTFOUND) if the requested frame is not found.'),
    ('gcpool', 'found' , 'Raise KeyError(VARIABLENOTFOUND) if the variable is not in the pool; ' +
                         'TypeError(WRONGDATATYPE) if it has the wrong type; ' +
                         'IndexError(INDEXOUTOFRANGE) if the start index is out of range.'),
    ('gdpool', 'found' , 'Raise KeyError(VARIABLENOTFOUND) if the variable is not in the pool; ' +
                         'TypeError(WRONGDATATYPE) if it has the wrong type; ' +
                         'IndexError(INDEXOUTOFRANGE) if the start index is out of range.'),
    ('gipool', 'found' , 'Raise KeyError(VARIABLENOTFOUND) if the variable is not in the pool; ' +
                         'TypeError(WRONGDATATYPE) if it has the wrong type; ' +
                         'IndexError(INDEXOUTOFRANGE) if the start index is out of range.'),
    ('gnpool', 'found' , 'Raise KeyError(VARIABLENOTFOUND) if the variable is not in the pool; ' +
                         'TypeError(WRONGDATATYPE) if it has the wrong type; ' +
                         'IndexError(INDEXOUTOFRANGE) if the start index is out of range.'),
    ('invert', ''      , 'Raise ValueError(SINGULARMATRIX) if the matrix is singular.'),
    ('namfrm', ''      , 'Raise KeyError(FRAMENAMENOTFOUND) if the frame name is not recognized.'),
    ('pckcov', ''      , 'Raise KeyError(FRAMEIDNOTFOUND) if the frame is not found.'),
    ('spkcov', ''      , 'Raise KeyError(BODYIDNOTFOUND) if the body is not found.'),
    ('srfc2s', ''      , 'Raise KeyError(NOTRANSLATION) if surface not found.'),
    ('srfcss', 'isname', 'Raise KeyError(NOTRANSLATION) if surface not found.'),
    ('srfs2c', 'found' , 'Raise KeyError(NOTRANSLATION) if surface not found.'),
    ('srfscc', 'found' , 'Raise KeyError(NOTRANSLATION) if surface not found.'),
    ('stpool', 'found' , 'Raise KeyError(VARIABLENOTFOUND) if the variable is not in the pool; ' +
                         'TypeError(WRONGDATATYPE) if it has the wrong type; ' +
                         'IndexError(INDEXOUTOFRANGE) if the start index is out of range.'),
    ('tparse', 'errmsg', 'Raise ValueError(INVALIDTIMESTRING) if the sample string is invalid.'),
    ('tpictr', 'errmsg', 'Raise ValueError(INVALIDTIMESTRING) if the sample string is invalid.'),

# Functions defined in cspyce0_part2.i
    ('ckfrot',  'found', 'Raise IOError(CKINSUFFDATA) if the requested information is unavailable.'),
    ('ckfxfm',  'found', 'Raise IOError(CKINSUFFDATA) if the requested information is unavailable.'),
    ('dafgsr',  'found', 'Raise IOError(DAFFRNOTFOUND) if the file read failed.'),
    ('dlabbs',  'found', 'Raise IOError(DASFILEREADFAILED) if the file read failed.'),
    ('dlabfs',  'found', 'Raise IOError(DASFILEREADFAILED) if the file read failed.'),
    ('dlafns',  'found', 'Raise IOError(DASFILEREADFAILED) if the file read failed.'),
    ('dlafps',  'found', 'Raise IOError(DASFILEREADFAILED) if the file read failed.'),
#     ('dnearp',  'found', 'Raise ValueError(DEGENERATESURFACE) if the point is degenerate.'),  removed, not really an error condition
    ('dskx02',  'found', 'Raise ValueError(NOINTERCEPT) if the intercept does not exist.'),
    ('dskxsi',  'found', 'Raise ValueError(NOINTERCEPT) if the intercept does not exist.'),
    ('ekfind', ('error',
              'errmsg'), 'Raise ValueError(INVALIDVALUE) if the query is invalid.'),
    ('ekgc'  ,  'found', 'Raise IndexError(INDEXOUTOFRANGE) if the column is not present in the row.'),
    ('ekgd'  ,  'found', 'Raise IndexError(INDEXOUTOFRANGE) if the column is not present in the row.'),
    ('ekgi'  ,  'found', 'Raise IndexError(INDEXOUTOFRANGE) if the column is not present in the row.'),
    ('ekpsel', ('error',
              'errmsg'), 'Raise ValueError(INVALIDVALUE) if the query is invalid.'),
    ('hx2dp' , ('error',
              'errmsg'), 'Raise ValueError(INVALIDVALUE) if the hex string is invalid.'),
    ('kdata' ,  'found', 'Raise IOError(FILENOTFOUND) if the specified file could be located.'),
    ('kinfo' ,  'found', 'Raise IOError(FILENOTFOUND) if the specified file could be located.'),
    ('spksfs',  'found', 'Raise IOError(SPKINSUFFDATA) if the requested information is unavailable.'),
    ('szpool',  'found', 'Raise KeyError(KERNELVARNOTFOUND) if the parameter name is not recognized.'),
    ('tkfram',  'found', 'Raise KeyError(FRAMEIDNOTFOUND) if the frame transformation could not be found.'),
]

# Add a new set of dictionary entries for each "_error" function
for (func, name, message) in ERROR_INFO:
    key = func + '_error'
    CSPYCE_ARGNAMES[key] = CSPYCE_ARGNAMES[func]
    CSPYCE_ABSTRACT[key] = CSPYCE_ABSTRACT[func]
    CSPYCE_URL[key] = CSPYCE_URL[func]

    if func in CSPYCE_DEFAULTS:
        CSPYCE_DEFAULTS[key] = CSPYCE_DEFAULTS[func]

    if name:
        CSPYCE_RETNAMES[key] = list(CSPYCE_RETNAMES[func])
        CSPYCE_DEFINITIONS[key] = CSPYCE_DEFINITIONS[func].copy()
        if isinstance(name, str):
            name = (name,)
        for n in name:
            CSPYCE_RETNAMES[key].remove(n)
            del(CSPYCE_DEFINITIONS[key][n])
    else:
        CSPYCE_RETNAMES[key] = CSPYCE_RETNAMES[func]
        CSPYCE_DEFINITIONS[key] = CSPYCE_DEFINITIONS[func]

    if func in CSPYCE_PS:
        CSPYCE_PS[key] = CSPYCE_PS[func] + '\n\n' + message
    else:
        CSPYCE_PS[key] = message

    old_defs = CSPYCE_DEFINITIONS[func]
    CSPYCE_SIGNATURES[key] = [old_defs[n][0] for n in CSPYCE_ARGNAMES[key]]
    CSPYCE_RETURNS[key]    = [old_defs[n][0] for n in CSPYCE_RETNAMES[key]]

################################################################################



#!/usr/bin/env python

# If you are installing this code via "pip install cspyce", then you should
# never see this file.  pip automatically determines whether you need a binary
# distribution or source distribution, and automatically builds it as needed for
# your machine.
#
# If you are doing a build from sources, please read README-developers.md in
# this directory.


from glob import glob
import os
import platform
import subprocess
import sys

from setuptools.command.build_ext import build_ext

from get_spice import GetCspice

# We prefer setuptools, but will use distutils if setuptools isn't available
try:
    from setuptools import Command, setup, Extension
except ImportError:
    from distutils.core import Command, setup, Extension

try:
    import numpy
except ImportError:
    subprocess.check_call([sys.executable, "-m", "pip", "install", "numpy"])
    import numpy


PYTHON2 = sys.version_info[0] < 3
IS_LINUX = platform.system() == "Linux"
IS_MACOS = platform.system() == "Darwin"
IS_WINDOWS = platform.system() == "Windows"
assert IS_LINUX or IS_MACOS or IS_WINDOWS


class GenerateCommand(Command):
    user_options = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        if not PYTHON2:
            from swig.make_vectorize import create_vectorize_header_file
            from swig.make_cspyce0_info import make_cspice0_info
            create_vectorize_header_file("swig/vectorize.i")
            make_cspice0_info("cspyce/cspyce0_info.py")
            command = "swig -python -outdir cspyce/. " \
                      "-o swig/cspyce0_wrap.c swig/cspyce0.i".split(" ")
            subprocess.check_call(command)
            command = "swig -python -outdir cspyce/. " \
                      "-o swig/typemap_samples_wrap.c swig/typemap_samples.i".split(" ")
            subprocess.check_call(command)
        else:
            command = "echo Cannot rebuild files in Python2".split(" ")
            subprocess.check_call(command)


# Some linkers seem to have trouble with 2400 files, so we break it up into
# smaller libraries with a maximum of 250 files apiece.

cspice_directory = GetCspice().download()

def get_c_libraries():
    files = sorted(glob(os.path.join(cspice_directory, "src", "*.c")))
    splits = 1 if IS_LINUX else 1 + (len(files) // 250)
    compiler_flags = ["-DKR_headers", "-DMSDOS", "/nowarn"] if IS_WINDOWS else ["-w"]
    cspice_libraries = [
        ("cspice_" + str(i + 1), {
            "sources": files[i::splits],
            "include_dirs": [os.path.join(cspice_directory, "include")],
            "cflags": compiler_flags
        })
        for i in range(splits)]
    return cspice_libraries


def get_extensions():
    cspyce_cflags = ["/nowarn", "/DSWIG_PYTHON_CAST_MODE"] if IS_WINDOWS \
        else ["-Wno-incompatible-pointer-types", "-DSWIG_PYTHON_CAST_MODE"]
    include_dirs = [os.path.join(cspice_directory, "include"), numpy.get_include()]

    cspyce0_module = Extension(
        "cspyce._cspyce0",
        sources=["swig/cspyce0_wrap.c"],
        include_dirs=include_dirs,
        extra_compile_args=cspyce_cflags)

    typemap_samples_module = Extension(
        "cspyce._typemap_samples",
        sources=["swig/typemap_samples_wrap.c"],
        include_dirs=include_dirs,
        extra_compile_args=cspyce_cflags,
    )
    return [cspyce0_module, typemap_samples_module]


class MyBuildExt(build_ext):
    def initialize_options(self):
        build_ext.initialize_options(self)


def do_setup():
    prerelease_version = os.getenv("PRERELEASE_VERSION", "")
    if prerelease_version == "release":
        prerelease_version = ""

    setup(
        version='2.0.9' + prerelease_version,
        ext_modules=get_extensions(),
        libraries=get_c_libraries(),
        cmdclass={
            "build_ext": MyBuildExt,
            "generate": GenerateCommand,
        }
    )

do_setup()
